/*----------------------------------------------------------------------
  File    : genpst.c
  Contents: generate parallel spike trains with injected coincidences
  Author  : Christian Borgelt
  History : 2008.06.17 file created
            2009.05.25 missing evaluation of option -g added
            2009.05.26 assembly sizes generalized
            2009.05.29 enforcing coincidence frequencies added
            2009.06.10 generalized, spike probability oscillation added
            2009.06.11 neuron index lists generalized from mere ranges
            2010.11.15 check of neuron ids against range added
            2011.07.22 end marker for gdf file made optional (-e)
            2011.08.18 enforcing expected number of copied spikes added
            2011.09.01 enforcing the gaps added (no background spikes)
            2011.10.08 bug concerning standard operation fixed (xfq)
            2012.02.04 bug in cycle length initialization fixed
            2013.08.09 adapted to direction parameter of int_heapsort()
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include "arrays.h"
#include "random.h"

#ifndef INFINITY
#define INFINITY    (DBL_MAX+DBL_MAX)
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "genpst"
#define DESCRIPTION "generate parallel spike trains " \
                    "with injected coincidences"
#define VERSION     "version 3.5 (2014.10.24)         " \
                    "(c) 2009-2014   Christian Borgelt"

/* --- sizes and constants --- */
#define BLKSIZE       16        /* block size for array enlargement */
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI          3.14159265358979323846  /* pi */
#endif

/* --- error codes --- */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* file open  failed */
#define E_FREAD      (-3)       /* file read  failed */
#define E_FWRITE     (-4)       /* file write failed (unused) */
#define E_OPTION     (-5)       /* unknown option */
#define E_OPTARG     (-6)       /* missing option argument */
#define E_ARGCNT     (-7)       /* too few/many arguments */
#define E_TIME       (-8)       /* invalid number of time bins */
#define E_COUNT      (-9)       /* invalid number of neurons */
#define E_FORMAT    (-10)       /* invalid parameter format */
#define E_VALUE     (-11)       /* invalid parameter value */
#define E_PARAM     (-12)       /* invalid parameter value */
#define E_NID       (-13)       /* invalid neuron identifier */
#define E_NEURON    (-14)       /* invalid parameters for neuron */
#define E_UNKNOWN   (-15)       /* unknown error */

#define NID(l,i)   (((l)->ids) ? (l)->ids[i]-1 : (l)->beg +(i)-1)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- a neuron --- */
  int    spike;                 /* flag for spike in current bin */
  double prob;                  /* background firing probability */
  double amp;                   /* prob. oscillation amplitude */
  double len;                   /* prob. oscillation cycle length */
  double phi;                   /* prob. oscillation cycle phase */
  double tmp;                   /* buffer for setup computations */
} NEURON;                       /* (neuron) */

typedef struct {                /* --- a list of neuron ids --- */
  int    beg;                   /* id of first neuron */
  int    end;                   /* id of last  neuron */
  int    size;                  /* size of neuron id array */
  int    cnt;                   /* number of neuron ids */
  int    *ids;                  /* array  of neuron ids */
} IDLIST;                       /* (list of neuron ids) */

typedef struct {                /* --- a neuron group --- */
  IDLIST list;                  /* list of neuron ids */
  double prob;                  /* (marginal) firing probability */
} GROUP;                        /* (neuron group) */

typedef struct {                /* --- a neuron assembly --- */
  IDLIST list;                  /* list of neuron ids */
  double prob;                  /* firing probability */
  double copy;                  /* copy   probability */
  double curr;                  /* current firing probability */
  double exp;                   /* expected number of coincidences */
  int    act;                   /* actual   number of coincidences */
  int    gap;                   /* whether to enforce the gaps */
  double xfq;                   /* expected spike frequency */
  int    *frqs;                 /* actual spike freqs. per neuron */
} ASSEMB;                       /* (neuron assembly) */

typedef struct {                /* --- a neuron oscillation group --- */
  IDLIST list;                  /* list of neuron ids */
  double amp;                   /* prob. oscillation amplitude */
  double len;                   /* prob. oscillation cycle length */
  double phi;                   /* prob. oscillation cycle phase */
} OSCILL;                       /* (neuron oscillation) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_OPTION   -5 */  "unknown option -%c",
  /* E_OPTARG   -6 */  "missing option argument",
  /* E_ARGCNT   -7 */  "wrong number of arguments",
  /* E_TIME     -8 */  "invalid number of time bins %d",
  /* E_COUNT    -9 */  "invalid number of neurons %d",
  /* E_FORMAT  -10 */  "invalid parameter format %s",
  /* E_VALUE   -11 */  "invalid parameter value in %s",
  /* E_PARAM   -12 */  "invalid parameter value %g",
  /* E_NID     -13 */  "invalid neuron identifier %d",
  /* E_NEURON  -14 */  "invalid parameters for neuron %d",
  /* E_UNKNOWN -15 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static const char *prgname;     /* program name for error messages */

static int    grpsize  = 0;     /* size of group array */
static int    grpcnt   = 0;     /* number of marg. probability groups */
static GROUP  *groups  = NULL;  /* neuron marg. probability groups */
static int    asmsize  = 0;     /* size of neuron assembly array */
static int    asmcnt   = 0;     /* number of neuron assemblies */
static ASSEMB *assembs = NULL;  /* neuron assemblies */
static int    oscsize  = 0;     /* size of oscillation array */
static int    osccnt   = 0;     /* number of oscillation groups */
static OSCILL *oscills = NULL;  /* neuron oscillation groups */
static NEURON *neurons = NULL;  /* neurons (simulation parameters) */
static FILE   *out     = NULL;  /* output file */

/*----------------------------------------------------------------------
  Parameter Parsing Functions
----------------------------------------------------------------------*/

static int expand (IDLIST *list)
{                               /* --- expand a neuron index range */
  int n;                        /* number of array elements needed */
  int *p;                       /* buffer for reallocation */

  assert(list);                 /* check the function arguments */
  n = list->cnt +(list->end -list->beg +1);
  if (n > list->size) {         /* if the index array is too small */
    p = (int*)realloc(list->ids, (size_t)(n += BLKSIZE) *sizeof(int));
    if (!p) return E_NOMEM;     /* enlarge the index array */
    list->ids = p; list->size = n;   /* set the new array */
  }                             /* and the new array size */
  for (n = list->beg; n <= list->end; n++)
    list->ids[list->cnt++] = n; /* store the neuron indices */
  return 0;                     /* return 'ok' */
}  /* expand() */

/*--------------------------------------------------------------------*/

static int get_ids (IDLIST *list, const char *s, char **e)
{                               /* --- get list of neuron indices */
  assert(list && s && e);       /* check the function arguments */
  list->ids = NULL;             /* clear the index array */
  list->cnt = list->size = 0;   /* and its size */
  while (1) {                   /* get index of (first) neuron */
    list->beg = (int)strtol(s, e, 0);
    if (*e == s)                 return E_FORMAT;
    if (list->beg <= 0)          return E_VALUE;
    if (**e != '-')             /* if no range of neurons, */
      list->end = list->beg;    /* set end index to start index */
    else {                      /* if range of neurons */
      list->end = (int)strtol(s = *e+1, e, 0);
      if (*e == s)               return E_FORMAT;
      if (list->end < list->beg) return E_VALUE;
    }                           /* get the index of the last neuron */
    if (**e != ',') break;      /* if end of neuron list, abort loop */
    s = *e +1;                  /* skip the separator (the comma) */
    if (expand(list) < 0)        return E_NOMEM;
  }                             /* expand the current index range */
  if (!list->ids)               /* if only a single index range, */
    list->cnt = list->end -list->beg +1;/* get number of neurons */
  else {                        /* expand the last neuron range */
    if (expand(list) < 0)        return E_NOMEM;
    int_heapsort(list->ids, (size_t)list->cnt, +1);
    list->cnt = (int)int_unique(list->ids, (size_t)list->cnt);
  }                             /* remove duplicate neuron indices */
  return list->cnt;             /* return the number of indices */
}  /* get_ids() */

/*--------------------------------------------------------------------*/

static int add_grp (const char *s, char **e)
{                               /* --- add a neuron group */
  int   r;                      /* buffer for result */
  GROUP *g;                     /* buffer for a group */

  assert(s && e);               /* check the function arguments */
  if (grpcnt >= grpsize) {      /* if the groups array is full */
    grpsize += (grpsize > BLKSIZE) ? grpsize >> 1 : BLKSIZE;
    g = (GROUP*)realloc(groups, (size_t)grpsize *sizeof(GROUP));
    if (!g) return E_NOMEM;     /* enlarge the groups array */
    groups = g;                 /* and set the new array */
  }
  g = groups +grpcnt;           /* get the next free group */
  r = get_ids(&g->list, s, e);  /* get indices of neurons in group */
  if (r < 0) return r;          /* check for a read error */
  if (**e != ':')                     return E_FORMAT;
  g->prob = strtod(s = *e+1, e);/* get firing probability */
  if (*e == s)                        return E_FORMAT;
  if ((g->prob < 0) || (g->prob > 1)) return E_VALUE;
  return ++grpcnt;              /* count the added group */
}  /* add_grp() */

/*--------------------------------------------------------------------*/

static int add_asm (const char *s, char **e)
{                               /* --- add a neuron assembly */
  int    r;                     /* buffer for result */
  ASSEMB *a;                    /* buffer for an assembly */

  assert(s && e);               /* check the function arguments */
  if (asmcnt >= asmsize) {      /* if the groups array is full */
    asmsize += (asmsize > BLKSIZE) ? asmsize >> 1 : BLKSIZE;
    a = (ASSEMB*)realloc(assembs, (size_t)asmsize *sizeof(ASSEMB));
    if (!a) return E_NOMEM;     /* enlarge the assemblies array */
    assembs = a;                /* and set the new array */
  }
  a = assembs +asmcnt;          /* get the next free assembly */
  r = get_ids(&a->list, s, e);  /* get indices of neurons in asm. */
  if (r < 0) return r;          /* check for a read error */
  if (**e != ':')          return E_FORMAT;
  a->prob = strtod(s = *e+1, e);/* get firing probability */
  if (*e == s)             return E_FORMAT;
  if (fabs(a->prob) > 1)   return E_VALUE;
  if (!**e) { a->copy = 1; a->xfq = 0; return ++asmcnt; }
  if (**e != ':')          return E_FORMAT;
  a->copy = strtod(s = *e+1, e);/* get spike copy probability */
  if (*e == s)             return E_FORMAT;
  if (fabs(a->copy) > 1)   return E_VALUE;
  if ((**e != '+') && (**e != '*'))
    a->xfq  = -1;               /* if no strict expectation requested */
  else {                        /* clear the expected frequency */
    a->gap  = (*(*e)++ == '*'); /* get flag for enforcing gaps */
    a->xfq  =  fabs(a->prob *a->copy);
    a->prob = -fabs(a->prob);   /* compute the expected frequency */
  }                             /* and mark the copy probability */
  return ++asmcnt;              /* count the added assembly */
}  /* add_asm() */

/*--------------------------------------------------------------------*/

static int add_osc (const char *s, char **e)
{                               /* --- add a neuron oscillation */
  int    r;                     /* buffer for result */
  OSCILL *o;                    /* buffer for an oscillation */

  assert(s && e);               /* check the function arguments */
  if (osccnt >= oscsize) {      /* if the groups array is full */
    oscsize += (oscsize > BLKSIZE) ? oscsize >> 1 : BLKSIZE;
    o = (OSCILL*)realloc(oscills, (size_t)oscsize *sizeof(OSCILL));
    if (!o) return E_NOMEM;     /* enlarge the oscillation array */
    oscills = o;                /* and set the new array */
  }
  o = oscills +osccnt;          /* get the next free oscillation */
  r = get_ids(&o->list, s, e);  /* get indices of neurons in osc. */
  if (r < 0) return r;          /* check for a read error */
  if (**e != ':')                   return E_FORMAT;
  o->amp = strtod(s = *e+1, e); /* get amplitude of oscillation */
  if (*e == s)                      return E_FORMAT;
  if ((o->amp < 0) || (o->amp > 1)) return E_VALUE;
  if (**e != ':')                   return E_FORMAT;
  o->len = strtod(s = *e+1, e); /* get cycle length of oscillation */
  if (*e == s)                      return E_FORMAT;
  if (o->len <= 0)                  return E_VALUE;
  if (!**e) { o->phi = 0; return ++osccnt; }
  if (**e != ':')                   return E_FORMAT;
  o->phi = strtod(s = *e+1, e); /* get phase of oscillation */
  if (*e == s)                      return E_FORMAT;
  return ++osccnt;              /* count the added oscillation */
}  /* add_osc() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print an error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
    fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print error message */
    va_end(args);               /* end argument evaluation */
    fputc('\n', stderr);        /* terminate the output line */
  }
  #ifndef NDEBUG                /* if debug version */
  if (neurons) free(neurons);   /* deallocate all memory */
  if (assembs) free(assembs);
  if (oscills) free(oscills);
  if (out && (out != stdout)) fclose(out);
  #endif                        /* close the output file */
  exit(code);                   /* abort the program */
}  /* error() */

/*--------------------------------------------------------------------*/

static int collect (ASSEMB *asm, double min, double max, int *buf)
{                               /* --- collect eligible neurons */
  int    i;                     /* loop variable */
  IDLIST *list;                 /* list of neuron ids */
  int    *p = buf;              /* to traverse the buffer */
  double d;                     /* difference to expectation */

  assert(asm && buf             /* check the function arguments */
  &&     asm->frqs && (asm->xfq >= 0));
  list = &asm->list;            /* traverse the assembly neurons */
  for (i = 0; i < list->cnt; i++) {
    d = asm->xfq -asm->frqs[i]; /* compute difference to expectation */
    if ((d >= min) && (d <= max)) *p++ = i;
  }                             /* collect qualifying neurons */
  return (int)(p -buf);         /* return the number of neuron ids */
}  /* collect() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int    i, f, x, k = 0;        /* loop variables */
  char   *s, *z;                /* to traverse the options */
  char   **optarg = NULL;       /* option argument */
  char   *fname   = NULL;       /* name of output gdf file */
  int    t        = 10000;      /* number of time steps */
  int    n        = 100;        /* total number of neurons */
  double p        = 0.020;      /* spike probability per time bin */
  int    gdf      = 0;          /* flag for gdf output */
  double delta    = 1.0;        /* length of a time bin   (for gdf) */
  int    end      = 1;          /* flag for an end marker (for gdf) */
  long   seed;                  /* seed value for random numbers */
  GROUP  *g;                    /* to traverse the neuron groups */
  ASSEMB *a;                    /* to traverse the assemblies */
  OSCILL *o;                    /* to traverse the oscillations */
  NEURON *u;                    /* to traverse the neurons */
  IDLIST *l;                    /* to access the neuron id lists */
  int    m, r;                  /* number of collected neurons */
  double exp;                   /* expected number of copied spikes */
  int    act;                   /* actual   number of copied spikes */
  int    *buf, *b;              /* buffer for neuron ids */

  prgname = argv[0];            /* get program name for error msgs. */

  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no arguments given */
    printf("usage: %s [options] file\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-t#    number of time bins                     "
                  "(default: %d)\n", t);
    printf("-n#    total number of neurons                 "
                  "(default: %d)\n", n);
    printf("-p#    default spike probability per time bin  "
                  "(default: %g)\n", p);
    printf("-q#2   spike probability for a group           "
                  "(default: none)\n");
    printf("       format: -ql:p\n");
    printf("           l - list of neuron indices\n");
    printf("           p - spike probability\n");
    printf("       format of neuron list: (a-z|i)[,(a-z|i)]*\n");
    printf("           a - index of first neuron in a range\n");
    printf("           z - index of last  neuron in a range\n");
    printf("           i - index of individual neuron\n");
    printf("-m#3   neuron assembly specification           "
                  "(default: none)\n");
    printf("       format: -ml:p[:c[+]]\n");
    printf("           l - list of neuron indices\n");
    printf("           p - firing probability\n");
    printf("           c - copy probability (optional)\n");
    printf("           + - enforce expected value per neuron\n");
    printf("       negative probabilities: enforce expected values\n");
    printf("-o#4   neuron oscillation specification        "
                  "(default: none)\n");
    printf("       format: -ol:a:c[:p]\n");
    printf("           l - list of neuron indices\n");
    printf("           a - amplitude (as fraction)\n");
    printf("           c - cycle length (in bins)\n");
    printf("           p - cycle phase  (in bins, optional)\n");
    printf("-S#    seed value for random numbers           "
                  "(default: time)\n");
    printf("-g     write output file as gdf                "
                  "(default: transactions)\n");
    printf("-d#    length of a time bin for gdf output     "
                  "(default: %g)\n", delta);
    printf("-e     do not write an end marker to gdf file  "
                  "(default: -1 marker)\n");
    printf("file   output file (\"-\": standard output)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get the default seed value */
  for (i = k = 0; ++i < argc; ){/* traverse the arguments */
    s = argv[i];                /* get the next argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        x = 0;                  /* clear the error flag */
        switch (*s++) {         /* evaluate switches */
          case 't': t = (int)strtol(s, &s, 0); break;
          case 'n': n = (int)strtol(s, &s, 0); break;
          case 'p': p =      strtod(s, &s);    break;
          case 'q': x = add_grp(z = s, &s);    break;
          case 'm': x = add_asm(z = s, &s);    break;
          case 'o': x = add_osc(z = s, &s);    break;
          case 'S': seed  =  strtol(s, &s, 0); break;
          case 'g': gdf   = -1;                break;
          case 'd': delta =  strtod(s, &s);    break;
          case 'e': end   =  0;                break;
          default : error(E_OPTION, *--s);     break;
        }                       /* set option variables */
        if (x < 0) error(x, z); /* check for an error */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fname = s;       break;
        default: error(E_ARGCNT); break;
      }                         /* note filename */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (k != 1) error(E_ARGCNT);  /* and number of arguments */
  if (t  <= 0)            error(E_TIME,  t);
  if (n  <= 0)            error(E_COUNT, n);
  if ((p < 0) || (p > 1)) error(E_PARAM, p);
  fprintf(stderr, "\n");        /* terminate the startup message */

  /* --- set up parameters and arrays --- */
  for (k = x = 0; k < asmcnt; k++) /* sum the needed frequencies */
    if (assembs[k].xfq > 0) x += assembs[k].list.cnt;
  neurons = (NEURON*)malloc((size_t)n       *sizeof(NEURON)
                           +(size_t)(n+n+x) *sizeof(int));
  if (!neurons) error(E_NOMEM); /* create the neuron array */
  buf = (int*)(neurons +n);     /* and the neuron id buffer */
  memset(b = buf+n+n, 0, (size_t)x *sizeof(int));
  for (i = 0; i < n; i++) {     /* traverse the neurons and */
    u = neurons +i;             /* initialize the parameters */
    u->spike = 0; u->prob = p;  /* with default values */
    u->phi = u->len = u->amp = 0; u->tmp = 1;
  }
  for (k = 0; k < grpcnt; k++){ /* traverse the neuron groups */
    g = groups+k; l = &g->list; /* and the neurons in each group */
    for (i = 0; i < l->cnt; i++) {
      if ((x = NID(l,i)) >= n) error(E_NID, x+1);
      neurons[x].prob = g->prob;
    }                           /* set the spike probability */
  }                             /* for the neurons in each group */
  for (k = 0; k < osccnt; k++){ /* traverse the oscillations */
    o = oscills+k; l = &o->list;/* and the neurons in each group */
    for (i = 0; i < l->cnt; i++) {
      if ((x = NID(l,i)) >= n)       error(E_NID, x+1);
      if ((u = neurons +x)->len > 0) error(E_NID, x+1);
      u->amp = o->amp; u->len = o->len; u->phi = o->phi;
    }                           /* set the oscillation parameters */
  }                             /* for each neuron in a group */
  for (k = 0; k < asmcnt; k++){ /* traverse the neuron assemblies */
    a = assembs+k; l = &a->list;/* and the neurons in each assembly */
    a->curr = fabs(a->prob);    /* init. the firing probability, */
    a->exp  = a->curr *t;       /* the expected coincidences */
    a->act  = 0;                /* and the actual coincidences */
    exp     = floor(a->exp +0.5);
    if (fabs(a->exp/exp)-1 < 1e-9) a->exp = exp;
    if (a->xfq < 0)             /* if random spike frequencies, */
      a->frqs = NULL;           /* no frequency counters are needed */
    else {                      /* if to enforce expectations, */
      a->frqs = b; b += l->cnt; /* store the frequency buffer */
      a->xfq *= t;              /* compute expected spike frequency */
      exp = floor(a->xfq +0.5);
      if (fabs(a->xfq/exp)-1 < 1e-9) a->xfq = exp;
    }                           /* round to integer if close */
    for (i = 0; i < l->cnt; i++) { /* traverse the neurons */
      if ((x = NID(l,i)) >= n) error(E_NID, x+1);
      if (a->gap) neurons[x].tmp *= 1 -a->curr;
      else        neurons[x].tmp *= 1 -a->curr *fabs(a->copy);
    }                           /* collect assembly parameters */
  }                             /* for the neurons */
  for (i = 0; i < n; i++) {     /* traverse the neurons */
    u = neurons +i; p = 1 -u->prob;
    if ((u->tmp <= 0) || (u->tmp < p)) error(E_NEURON, i+1);
    u->prob = 1 -p /u->tmp;     /* set background spike probability */
    u->amp *= (p < u->prob) ? p : u->prob;
    u->len /= 2*M_PI;           /* scale the oscillation amplitude */
  }                             /* and the oscillation cycle length */

  /* --- write output file --- */
  if (!*fname) { out = stdout; fname = "<stdout>"; }
  else         { out = fopen(fname, "w"); }
  if (!out) error(E_FOPEN, fname);     /* open the file for writing */
  fprintf(stderr, "writing %s ... ", fname);
  rseed((unsigned)seed);        /* seed the random number generator */
  for (x = 0; x < t; x++) {     /* traverse the time bins */
    for (i = 0; i < n; i++)     /* traverse the neurons */
      neurons[i].spike = 0;     /* and clear all spikes */
    for (k = 0; k < asmcnt; k++) {
      a = assembs+k; l = &a->list;   /* traverse neuron assemblies */
      if (a->prob < 0)          /* if to enforce exp. coincidences */
        a->curr = (a->exp -a->act) /(t-x);
      if (drand() >= a->curr) continue;
      p   = fabs(a->copy);      /* get the copy probability */
      exp = l->cnt *p;          /* and the expected number of neurons */
      act = 0;                  /* and init. actual number of neurons */
      if (a->xfq <= 0) {        /* if random spike frequencies */
        for (i = 0; i < l->cnt; i++) { /* traverse the neurons */
          if (a->copy < 0)      /* if to enforce the neuron fraction */
            p = (exp -act) /(l->cnt -i);
          if (drand() >= p) continue;
          neurons[NID(l,i)].spike = 1;
          act++;                /* copy spike with current prob. */
        } }                     /* and count it for the coincidence */
      else {                    /* if to enforce spike expectation */
        r = collect(a, 1, INFINITY, buf);
        for (i = m = 0; i < r; i++) { /* traverse collected neurons */
          if (floor(a->xfq) -a->frqs[buf[i]] < floor(a->exp) -a->act)
            buf[m++] = buf[i];  /* collect optional neurons */
          else {                /* if a neuron is mandatory */
            neurons[NID(l,buf[i])].spike = 1;
            a->frqs[buf[i]]++;  /* set spike for the neuron and */
            act++;              /* count spike for the neuron */
          }                     /* and for the coincidence */
        }
        if (act+m < floor(exp)){/* if there are too few neurons */
          r = m; m = collect(a, DBL_MIN, 1-DBL_EPSILON, buf+n);
          for (i = 0; i < r; i++) {   /* traverse collected neurons */
            neurons[NID(l,buf[i])].spike = 1;
            a->frqs[buf[i]]++;  /* set spike for the neuron and */
            act++;              /* count spike for the neuron */
          }                     /* and for the coincidence */
          memcpy(buf, buf+n, (size_t)m *sizeof(int));
        }                       /* collect secondary eligible neurons */
        for (i = 0; i < m; i++){/* traverse the collected neurons */
          if (a->copy < 0)      /* if to enforce the neuron fraction */
            p = (exp -act) /(m -i);
          if (drand() >= p) continue;
          neurons[NID(l,buf[i])].spike = 1;
          a->frqs[buf[i]]++;    /* copy spike with current prob. */
          act++;                /* count spike for the neuron */
        }                       /* and for the coincidence */
        if (a->gap) {           /* if to enforce the gaps */
          for (i = 0; i < l->cnt; i++) {
            u = neurons +NID(l,i);
            if (u->spike <= 0) u->spike = -1;
          }                     /* traverse the neurons and */
        }                       /* mark the gaps in the group */
      }
      a->act++;                 /* count the created coincidence */
    }
    for (i = f = 0; i < n; i++){/* print the neuron spikes */
      if (neurons[i].spike < 0) /* if this is a marked gap, */
        continue;               /* definitely leave it empty */
      if (!neurons[i].spike) {  /* if there is no spike yet */
        u = neurons +i; p = u->prob;
        if (u->len > 0) p += u->amp *sin((u->phi +x) /u->len);
        if (drand() >= p) continue;
      }                         /* check for a background spike */
      if (gdf) fprintf(out, "%d %g\n", i+1, (x+0.5) *delta);
      else { if (f) fputc(' ', out); fprintf(out, "%d", f = i+1); }
    }                           /* print the neuron identifiers */
    if (!gdf) fprintf(out, "\n");
  }                             /* terminate the identifier list */
  if (gdf && end) fprintf(out, "-1 %g\n", t *delta);
  if (out != stdout) {          /* close the output file */
    k = fclose(out); if (k != 0) error(E_FWRITE, fname); }
  fprintf(stderr, "[%d neuron(s)/%d bin(s)] done.\n", n, t);

  /* --- clean up --- */
  #ifndef NDEBUG                /* if debug version */
  if (neurons) free(neurons);   /* deallocate all memory */
  if (assembs) free(assembs);
  if (oscills) free(oscills);
  #endif
  return 0;                     /* terminate the program */
}  /* main() */
